@@ -11,11 +11,12 @@ import os
 import sys
 import types
 
-import py
+import pathlib
 import pytest
 
 from .django_compat import is_django_unittest  # noqa
 from .fixtures import django_assert_num_queries  # noqa
+from .fixtures import django_assert_max_num_queries  # noqa
 from .fixtures import django_db_setup  # noqa
 from .fixtures import django_db_use_migrations  # noqa
 from .fixtures import django_db_keepdb  # noqa
@@ -30,6 +31,7 @@ from .fixtures import db  # noqa
 from .fixtures import django_user_model  # noqa
 from .fixtures import django_username_field  # noqa
 from .fixtures import live_server  # noqa
+from .fixtures import django_db_reset_sequences  # noqa
 from .fixtures import rf  # noqa
 from .fixtures import settings  # noqa
 from .fixtures import transactional_db  # noqa
@@ -87,13 +89,6 @@ def pytest_addoption(parser):
                   type='bool', default=False)
 
 
-def _exists(path, ignore=EnvironmentError):
-    try:
-        return path.check()
-    except ignore:
-        return False
-
-
 PROJECT_FOUND = ('pytest-django found a Django project in %s '
                  '(it contains manage.py) and added it to the Python path.\n'
                  'If this is wrong, add "django_find_project = false" to '
@@ -120,21 +115,28 @@ def _handle_import_error(extra_message):
 
 
 def _add_django_project_to_path(args):
-    args = [x for x in args if not str(x).startswith("-")]
+    def is_django_project(path):
+        return path.is_dir() and (path / 'manage.py').exists()
 
-    if not args:
-        args = [py.path.local()]
+    def find_django_path(args):
+        args = [pathlib.Path(x) for x in args if not str(x).startswith("-")]
+        args = [p for p in args if p.is_dir()]
 
-    for arg in args:
-        arg = py.path.local(arg)
+        if not args:
+            args = [pathlib.Path.cwd()]
 
-        for base in arg.parts(reverse=True):
-            manage_py_try = base.join('manage.py')
-
-            if _exists(manage_py_try):
-                sys.path.insert(0, str(base))
-                return PROJECT_FOUND % base
+        for arg in args:
+            if is_django_project(arg):
+                return arg
+            for parent in arg.parents:
+                if is_django_project(parent):
+                    return parent
+        return None
 
+    project_dir = find_django_path(args)
+    if project_dir:
+        sys.path.insert(0, str(project_dir))
+        return PROJECT_FOUND % project_dir
     return PROJECT_NOT_FOUND
 
 
@@ -148,7 +150,10 @@ def _setup_django():
     if not django.conf.settings.configured:
         return
 
-    django.setup()
+    import django.apps
+    if not django.apps.apps.ready:
+        django.setup()
+
     _blocking_manager.block()
 
 
@@ -183,6 +188,10 @@ def pytest_load_initial_conftests(early_config, parser, args):
         'the `urls` attribute of Django `TestCase` objects.  *modstr* is '
         'a string specifying the module of a URL config, e.g. '
         '"my_app.test_urls".')
+    early_config.addinivalue_line(
+        'markers',
+        'ignore_template_errors(): ignore errors from invalid template '
+        'variables (if --fail-on-template-vars is used).')
 
     options = parser.parse_known_args(args)
 
@@ -264,12 +273,12 @@ def _classmethod_is_defined_at_leaf(cls, method_name):
     super_method = None
 
     for base_cls in cls.__mro__[1:]:  # pragma: no branch
-        if hasattr(base_cls, method_name):
-            super_method = getattr(base_cls, method_name)
+        super_method = base_cls.__dict__.get(method_name)
+        if super_method is not None:
             break
 
     assert super_method is not None, (
-        '%s could not be found in base class' % method_name)
+        '%s could not be found in base classes' % method_name)
 
     method = getattr(cls, method_name)
 
@@ -293,7 +302,7 @@ def _disable_class_methods(cls):
 
     _disabled_classmethods[cls] = (
         # Get the classmethod object (not the resulting bound method),
-        # otherwise inheritence will be broken when restoring.
+        # otherwise inheritance will be broken when restoring.
         cls.__dict__.get('setUpClass'),
         _classmethod_is_defined_at_leaf(cls, 'setUpClass'),
         cls.__dict__.get('tearDownClass'),
@@ -379,13 +388,15 @@ def django_db_blocker():
 def _django_db_marker(request):
     """Implement the django_db marker, internal to pytest-django.
 
-    This will dynamically request the ``db`` or ``transactional_db``
-    fixtures as required by the django_db marker.
+    This will dynamically request the ``db``, ``transactional_db`` or
+    ``django_db_reset_sequences`` fixtures as required by the django_db marker.
     """
     marker = request.node.get_closest_marker('django_db')
     if marker:
-        transaction = validate_django_db(marker)
-        if transaction:
+        transaction, reset_sequences = validate_django_db(marker)
+        if reset_sequences:
+            request.getfixturevalue('django_db_reset_sequences')
+        elif transaction:
             request.getfixturevalue('transactional_db')
         else:
             request.getfixturevalue('db')
@@ -500,7 +511,7 @@ def _fail_for_invalid_template_variable(request):
     It does not raise an exception, but fails, as the stack trace doesn't
     offer any helpful information to debug.
     This behavior can be switched off using the marker:
-    ``ignore_template_errors``
+    ``pytest.mark.ignore_template_errors``
     """
     class InvalidVarException(object):
         """Custom handler for invalid strings in templates."""
@@ -667,11 +678,14 @@ _blocking_manager = _DatabaseBlocker()
 def validate_django_db(marker):
     """Validate the django_db marker.
 
-    It checks the signature and creates the `transaction` attribute on
-    the marker which will have the correct value.
+    It checks the signature and creates the ``transaction`` and
+    ``reset_sequences`` attributes on the marker which will have the
+    correct values.
+
+    A sequence reset is only allowed when combined with a transaction.
     """
-    def apifun(transaction=False):
-        return transaction
+    def apifun(transaction=False, reset_sequences=False):
+        return transaction, reset_sequences
     return apifun(*marker.args, **marker.kwargs)
 
 
